from typing import Tuple, List, Callable

import pandas as pd

from pandas_ml_common import get_pandas_object
from pandas_ml_common.utils import intersection_of_index, loc_if_not_none
from pandas_ml_common.utils.callable_utils import call_if_not_none


def extract_feature_labels_weights(
        df: pd.DataFrame,
        features_and_labels,
        **kwargs) -> Tuple[Tuple[pd.DataFrame, int], pd.DataFrame, pd.DataFrame, pd.Series, pd.Series]:
    features = get_pandas_object(df, features_and_labels.features, **kwargs).dropna()
    labels = get_pandas_object(df, features_and_labels.labels, **kwargs).dropna()
    targets = call_if_not_none(get_pandas_object(df, features_and_labels.targets, **kwargs), 'dropna')
    sample_weights = call_if_not_none(get_pandas_object(df, features_and_labels.sample_weights, **kwargs), 'dropna')
    gross_loss = call_if_not_none(get_pandas_object(df, features_and_labels.gross_loss, **kwargs), 'dropna')
    common_index = intersection_of_index(features, labels, targets, sample_weights)

    if features_and_labels.label_type is not None:
        labels = labels.astype(features_and_labels.label_type)

    return (
        (features.loc[common_index], len(df) - len(features) + 1),
        labels.loc[common_index],
        loc_if_not_none(targets, common_index),
        loc_if_not_none(sample_weights, common_index),
        loc_if_not_none(gross_loss, common_index)
    )


def extract_features(df: pd.DataFrame, features_and_labels, **kwargs) -> Tuple[List, pd.DataFrame, pd.DataFrame]:
    features = get_pandas_object(df, features_and_labels.features, **kwargs).dropna()
    targets = call_if_not_none(get_pandas_object(df, features_and_labels.targets, **kwargs), 'dropna')
    common_index = intersection_of_index(features, targets)

    if len(features) <= 0:
        raise ValueError("not enough data!")

    return (
        features_and_labels.label_columns,
        features.loc[common_index],
        loc_if_not_none(targets, common_index)
    )


def extract(features_and_labels, df, extractor, *args, **kwargs):
    return features_and_labels(df, extractor, *args, **kwargs)


def extract_with_post_processor(list, postprocessor: Callable):
    def extractor(df, **kwargs):
        extraction = get_pandas_object(df, list, **kwargs)
        return get_pandas_object(extraction, postprocessor, **kwargs)

    return extractor




import logging
import math
from copy import deepcopy
from typing import List, Callable

import numpy as np

from pandas_ml_common import Typing
from pandas_ml_utils.ml.summary import Summary
from .base_model import Model
from ..data.splitting.sampeling import Sampler
from ..data.splitting.sampeling.extract_multi_model_label import ExtractMultiMultiModelSampler

_log = logging.getLogger(__name__)


class MultiModel(Model):

    def __init__(self,
                 basis_model: Model,
                 nr_models: int,
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 **kwargs):
        super().__init__(basis_model.features_and_labels, summary_provider, **kwargs)
        self.basis_model = basis_model
        self.nr_models = nr_models
        self.sub_models: List[Model] = []

    def fit(self, sampler: Sampler, **kwargs) -> float:
        sub_model_losses = []

        # copy sampler and add a new cross validator extracting the model
        for i_model in range(self.nr_models):
            if "verbose" in kwargs and kwargs["verbose"]:
                print(f"fit model {i_model + 1} / {self.nr_models}")
            else:
                _log.info(f"fit model {i_model + 1} / {self.nr_models}")

            sm_loss = self.sub_models[i_model]\
                .fit(ExtractMultiMultiModelSampler(i_model, self.nr_models, sampler), **kwargs)
            sub_model_losses.append(sm_loss)

        self._history = [sm._history for sm in self.sub_models]
        return np.array(sub_model_losses).mean()

    def predict_sample(self, x: np.ndarray, **kwargs) -> np.ndarray:
        # call each sub_model.predict()
        predictions = [m.predict_sample(x) for m in self.sub_models]

        # and then concatenate the arrays back into one prediction
        return np.stack(predictions, 1)

    def plot_loss(self, figsize=(8, 6), columns=None):
        # override the plot function and plot the loss per model and fold
        import matplotlib.pyplot as plt

        columns = self.nr_models if columns is None else int(columns)
        rows = math.ceil(self.nr_models / columns)
        fig, axes = plt.subplots(rows, columns, figsize=(figsize if figsize else plt.rcParams.get('figure.figsize')))
        axes = axes.flatten()

        for m_nr, m_hist in enumerate(self._history):
            ax = axes[m_nr]

            for fold_nr, fold_loss in enumerate(m_hist):
                p = ax.plot(fold_loss[0], '-', label=f'{m_nr}: {fold_nr}: loss')
                ax.plot(fold_loss[1], '--', color=p[-1].get_color(), label=f'{m_nr}: {fold_nr}: val loss')

        plt.legend(loc='upper right')
        return fig, ax

    def __call__(self, *args, **kwargs):
        copy = deepcopy(self)
        nr_models = self.nr_models
        
        def create_sub_model(i):
            l = self.features_and_labels.labels
            w = self.features_and_labels.sample_weights

            fl = deepcopy(self.features_and_labels)
            fl._labels = [l[i*nr_models:(i+1)*nr_models]]
            fl._sample_weights = [w[i*nr_models:(i+1)*nr_models]] if w is not None else None

            sm = self.basis_model(*args, **kwargs)
            sm._features_and_labels = fl
            return sm

        copy.sub_models = [create_sub_model(i) for i in range(self.nr_models)]
        return copy



from unittest import TestCase

from sklearn.neural_network import MLPClassifier

from pandas_ml_common import np, pd
from pandas_ml_utils import SkModel, FeaturesAndLabels, MultiModel
from pandas_ml_utils.constants import PREDICTION_COLUMN_NAME
from pandas_ml_utils.ml.data.splitting import NaiveSplitter


class TestSkMultiModel(TestCase):

    def test_multi_model(self):
        """given some toy classification data"""
        df = pd.DataFrame({
            "a": [1, 0, 1, 0, 1, 0, 1, 0,],
            "b": [0, 0, 1, 1, 0, 0, 1, 1,],
            "c": [1, 0, 0, 1, 1, 0, 0, 1,],
            "d": [1, 0, 0, 1, 1, 0, 0, 1,],
        })

        model = MultiModel(
            SkModel(
                MLPClassifier(activation='logistic', max_iter=1000, hidden_layer_sizes=(3,), alpha=0.001,
                              solver='lbfgs', random_state=42),
                FeaturesAndLabels(features=["a", "b"], labels=["c", "d"], label_type=int),
            ),
            2
        )

        fit = df.model.fit(model, NaiveSplitter(0.49), epochs=1500, verbose=True)
        print(fit.training_summary.df)

        self.assertEqual(4, len(fit.training_summary.df[PREDICTION_COLUMN_NAME, "c"]))
        np.testing.assert_array_almost_equal(
            fit.training_summary.df[PREDICTION_COLUMN_NAME, "c"],
            fit.training_summary.df[PREDICTION_COLUMN_NAME, "d"]
        )

